#include "V4L2Device.hpp"

#include <cerrno>
#include <cstring>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>

using Buffer = V4L2Device::Buffer;

V4L2Device::V4L2Device(std::string& path)
{
    m_fd = ::open(path.c_str(), O_RDWR | O_NONBLOCK);
    if (m_fd == -1) {
        // Can't open device
        m_lastError = errno;
    } else {
        if (ioctl(VIDIOC_QUERYCAP, &m_caps)) {
            // Not V4L2 device
            ::close(m_fd);
            m_fd == -1;
        }
    }
}

V4L2Device::~V4L2Device()
{
    if (m_fd != -1)
        ::close(m_fd);
}

bool V4L2Device::isOpen() const { return m_fd != -1; }

bool V4L2Device::isError() const { return (m_fd == -1) || (m_lastError != 0); }

int V4L2Device::error() const { return m_lastError; }

int V4L2Device::fd() const { return m_fd; }

std::string V4L2Device::errorString() const { return std::string(strerror(m_lastError)); }

v4l2_buf_type V4L2Device::type() const
{
    if (m_caps.device_caps & V4L2_CAP_VIDEO_CAPTURE)
        return V4L2_BUF_TYPE_VIDEO_CAPTURE;
    else if (m_caps.device_caps & V4L2_CAP_VIDEO_OUTPUT)
        return V4L2_BUF_TYPE_VIDEO_OUTPUT;
    else if (m_caps.device_caps & V4L2_CAP_VBI_CAPTURE)
        return V4L2_BUF_TYPE_VBI_CAPTURE;
    else if (m_caps.device_caps & V4L2_CAP_VBI_OUTPUT)
        return V4L2_BUF_TYPE_VBI_OUTPUT;
    else
        return (v4l2_buf_type)0;
}

std::string V4L2Device::name() const { return std::string((const char*)m_caps.card); }

int V4L2Device::ioctl(int ctl, void* param)
{
    int result = ::ioctl(m_fd, ctl, param);
    m_lastError = (result < 0) ? errno : 0;
    return result;
}

std::vector<Buffer> V4L2Device::requestBuffers(int count)
{
    std::vector<Buffer> result;

    v4l2_requestbuffers reqbuf;
    reqbuf.type = type();
    reqbuf.count = count;
    reqbuf.memory = V4L2_MEMORY_MMAP;

    if (ioctl(VIDIOC_REQBUFS, &reqbuf)) {
        return result;
    }

    result.resize(reqbuf.count);

    for (int i = 0; i < (int)reqbuf.count; i++) {
        struct v4l2_buffer buffer;
        memset(&buffer, 0, sizeof(buffer));

        buffer.type = reqbuf.type;
        buffer.memory = V4L2_MEMORY_MMAP;
        buffer.index = i;

        if (ioctl(VIDIOC_QUERYBUF, &buffer))
            continue;

        result[i].index = i;
        result[i].length = buffer.length;
        result[i].data = (uint8_t*)mmap(
            NULL, buffer.length, PROT_READ | PROT_WRITE, MAP_SHARED, m_fd, buffer.m.offset);

        if (result[i].data == MAP_FAILED) {
            result[i].length = 0;
            result[i].data = nullptr;
        }
    }

    return result;
}

void V4L2Device::freeBuffers(std::vector<Buffer>& buffers)
{
    for (auto buf : buffers) {
        if (buf.data && buf.length)
            munmap(buf.data, buf.length);
    }

    v4l2_requestbuffers reqbuf;
    reqbuf.type = type();
    reqbuf.count = 0;
    reqbuf.memory = V4L2_MEMORY_MMAP;

    ioctl(VIDIOC_REQBUFS, &reqbuf);

    buffers.clear();
}
