#include "Board.hpp"
#include "Utils.hpp"
#include "V4L2Device.hpp"

#include <csignal>
#include <cstring>
#include <iomanip>
#include <iostream>
#include <poll.h>
#include <queue>

/*!
 * \dir ./
 * This example shows simple video passthrough example
 */

namespace {
static const int OUT_BUFFER_COUNT = 4;
static const int IN_BUFFER_COUNT = 8;
}

static bool shouldStop = false;

void stopSignal(int sig)
{
    if ((sig == SIGTERM) || (sig == SIGINT))
        shouldStop = true;
}

v4l2_format getFormat(V4L2Device& dev)
{
    v4l2_format fmt;
    fmt.type = dev.type();
    if (dev.ioctl(VIDIOC_G_FMT, &fmt)) {
        std::cerr << "Cannot get video format: " << dev.errorString() << std::endl;
        return v4l2_format();
    }
    return fmt;
}

v4l2_format setFormat(V4L2Device& dev, const v4l2_format& fmt)
{
    v4l2_format newFmt = fmt;
    if (dev.ioctl(VIDIOC_S_FMT, &newFmt)) {
        std::cerr << "Cannot set video format: " << dev.errorString() << std::endl;
        return fmt;
    }
    return newFmt;
}

v4l2_format setupFormat(V4L2Device& dev)
{
    // Get current frame format
    v4l2_format fmt = getFormat(dev);

    // Set field interleave - both fields are in one buffer
    // fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
    // Each field are in separate buffer
    fmt.fmt.pix.field = V4L2_FIELD_ALTERNATE;

    // 8-bit UYVY
    fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_UYVY;
    // 8-bit YUYV
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
    // 10-bit v210
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_V210;
    // Raw - whole SDI frame, with HANC, VANC, 10-bit dense packed
    // fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_SL_RAW;

    // Set and adjust format
    fmt = setFormat(dev, fmt);

    return fmt;
}

void preroll(
    V4L2Device& dev, const std::vector<V4L2Device::Buffer>& buffers, const v4l2_pix_format& fmt)
{
    // Fill playback queue
    struct v4l2_buffer v4l2buf;
    memset(&v4l2buf, 0, sizeof(v4l2buf));
    v4l2buf.type = dev.type();
    v4l2buf.memory = V4L2_MEMORY_MMAP;
    v4l2buf.field = V4L2_FIELD_ANY;

    for (auto buf : buffers) {
        v4l2buf.index = buf.index;
        if ((v4l2buf.type == V4L2_BUF_TYPE_VIDEO_OUTPUT) && (fmt.field == V4L2_FIELD_ALTERNATE))
            v4l2buf.field = (v4l2buf.index % 2) ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP;

        dev.ioctl(VIDIOC_QBUF, &v4l2buf);
    }
}

void mainLoop(V4L2Device& in, V4L2Device& out, std::vector<V4L2Device::Buffer>& inBuffers,
    std::vector<V4L2Device::Buffer>& outBuffers, const v4l2_pix_format& fmt)
{
    int inType = in.type();
    int outType = out.type();

    enum v4l2_field lastOutField;
    struct v4l2_buffer inV4L2Buf, outV4L2Buf;
    memset(&inV4L2Buf, 0, sizeof(inV4L2Buf));
    inV4L2Buf.type = inType;
    inV4L2Buf.memory = V4L2_MEMORY_MMAP;
    inV4L2Buf.field = V4L2_FIELD_ANY;
    memset(&outV4L2Buf, 0, sizeof(outV4L2Buf));
    outV4L2Buf.type = outType;
    outV4L2Buf.memory = V4L2_MEMORY_MMAP;

    // Start capture
    in.ioctl(VIDIOC_STREAMON, &inType);
    // Start playback
    out.ioctl(VIDIOC_STREAMON, &outType);

    std::queue<v4l2_buffer> inQueue, outQueue;

    pollfd pollfd[2];
    pollfd[0].fd = in.fd();
    pollfd[0].events = POLLIN;
    pollfd[1].fd = out.fd();
    pollfd[1].events = POLLOUT;

    while (!shouldStop) {
        // Wait for frame
        poll(pollfd, 2, 100);

        if (pollfd[0].revents & (POLLERR | POLLHUP | POLLNVAL))
            break;

        if (pollfd[1].revents & (POLLERR | POLLHUP | POLLNVAL))
            break;

        timespec timestamp;
        clock_gettime(CLOCK_MONOTONIC, &timestamp);

        // Buffer received
        if (pollfd[0].revents & POLLIN) {
            // Get in buffer from driver queue
            in.ioctl(VIDIOC_DQBUF, &inV4L2Buf);

            // Place it into our queue
            inQueue.push(inV4L2Buf);

            std::cout << "Received buffer at "
                      << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                      << "ns, id = " << inV4L2Buf.index << ", sequence = " << inV4L2Buf.sequence
                      << ", field = " << inV4L2Buf.field << ", timestamp = "
                      << (uint64_t)inV4L2Buf.timestamp.tv_sec * 1000000000ULL
                    + inV4L2Buf.timestamp.tv_usec * 1000
                      << "ns, size = " << inV4L2Buf.bytesused << std::endl;
        }

        // Buffer sended
        if (pollfd[1].revents & POLLOUT) {
            // Get sended buffer from queue
            out.ioctl(VIDIOC_DQBUF, &outV4L2Buf);

            // Place it into our queue
            outQueue.push(outV4L2Buf);

            std::cout << "Sended buffer at "
                      << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                      << "ns, id = " << inV4L2Buf.index << ", sequence = " << inV4L2Buf.sequence
                      << ", field = " << inV4L2Buf.field << ", timestamp = "
                      << (uint64_t)inV4L2Buf.timestamp.tv_sec * 1000000000ULL
                    + inV4L2Buf.timestamp.tv_usec * 1000
                      << "ns, size = " << inV4L2Buf.bytesused << std::endl;
        }

        // Has free output and free input buffers
        if (!inQueue.empty() && !outQueue.empty()) {
            outV4L2Buf = outQueue.front();
            outQueue.pop();

            inV4L2Buf = inQueue.front();
            inQueue.pop();

            V4L2Device::Buffer& dib = inBuffers[inV4L2Buf.index];
            V4L2Device::Buffer& dob = outBuffers[outV4L2Buf.index];

            /*
             * PROCESS BUFFER DATA HERE
             */
            // Copy input to output
            outV4L2Buf.field = inV4L2Buf.field;
            lastOutField = (enum v4l2_field)inV4L2Buf.field;
            memcpy(dob.data, dib.data, std::min(dob.length, dib.length));

            std::cout << "Sending buffers at "
                      << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                      << "ns, in id = " << inV4L2Buf.index << ", "
                      << "out id = " << outV4L2Buf.index << ", inQueue = " << inQueue.size()
                      << ", outQueue = " << outQueue.size() << std::endl;

            out.ioctl(VIDIOC_QBUF, &outV4L2Buf);
            in.ioctl(VIDIOC_QBUF, &inV4L2Buf);
        }

        if (inQueue.size() >= IN_BUFFER_COUNT / 2) {
            // Input is faster or there is frame drops on output, flush input queue
            std::cout << "Input queue overflow at "
                      << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                      << "ns, flushing" << std::endl;
            while (!inQueue.empty()) {
                inV4L2Buf = inQueue.front();
                in.ioctl(VIDIOC_QBUF, &inV4L2Buf);
                inQueue.pop();
            }
        }

        if (outQueue.size() >= OUT_BUFFER_COUNT / 2) {
            // Output is faster or there is frame drops on input, flush output queue
            std::cout << "Output queue overflow at "
                      << (uint64_t)timestamp.tv_sec * 1000000000ULL + timestamp.tv_nsec
                      << "ns, flushing" << std::endl;
            while (!outQueue.empty()) {
                outV4L2Buf = outQueue.front();
                outV4L2Buf.field = (fmt.field == V4L2_FIELD_ALTERNATE)
                    ? ((lastOutField == V4L2_FIELD_TOP) ? V4L2_FIELD_BOTTOM : V4L2_FIELD_TOP)
                    : fmt.field;
                lastOutField = (enum v4l2_field)outV4L2Buf.field;
                out.ioctl(VIDIOC_QBUF, &outV4L2Buf);
                outQueue.pop();
            }
        }
    }

    // Stop playback
    out.ioctl(VIDIOC_STREAMOFF, &outType);
    // Stop capture
    in.ioctl(VIDIOC_STREAMOFF, &inType);
}

int main(int argc, char** argv)
{
    std::string inDevPath, outDevPath;
    if (argc != 3) {
        std::cout << "Usage: " << argv[0] << " in_device out_device" << std::endl;
        return 0;
    }

    inDevPath = argv[1];
    std::cout << "Openning " << inDevPath << std::endl;
    V4L2Device in(inDevPath);

    if (!in.isOpen()) {
        std::cerr << "Cannot open input device: " << in.errorString() << std::endl;
        return -1;
    }

    if (in.type() != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
        std::cerr << "Device (" << in.name() << ") is not an input device" << std::endl;
        return -1;
    }

    outDevPath = argv[2];
    std::cout << "Openning " << outDevPath << std::endl;
    V4L2Device out(outDevPath);

    if (!out.isOpen()) {
        std::cerr << "Cannot open output device: " << out.errorString() << std::endl;
        return -1;
    }

    if (out.type() != V4L2_BUF_TYPE_VIDEO_OUTPUT) {
        std::cerr << "Device (" << out.name() << ") is not an output device" << std::endl;
        return -1;
    }

    signal(SIGTERM, &stopSignal);
    signal(SIGINT, &stopSignal);

    std::vector<V4L2Device::Buffer> inBuffers, outBuffers;
    v4l2_format format;

    setupFormat(out);
    format = setupFormat(in);

    inBuffers = in.requestBuffers(IN_BUFFER_COUNT);
    outBuffers = out.requestBuffers(OUT_BUFFER_COUNT);

    std::cout << "Streaming with " << inBuffers.size() << " input and " << outBuffers.size()
              << " buffers" << std::endl;

    preroll(in, inBuffers, format.fmt.pix);
    preroll(out, outBuffers, format.fmt.pix);

    mainLoop(in, out, inBuffers, outBuffers, format.fmt.pix);

    in.freeBuffers(inBuffers);
    out.freeBuffers(outBuffers);

    return 0;
}
